@@ -17,16 +17,29 @@ import copy
 
 import netaddr
 from novaclient import exceptions as nova_exception
+import six
 
-from ec2api.api import clients
 from ec2api.api import common
 from ec2api.api import ec2utils
+from ec2api.api import vpn_connection as vpn_connection_api
+from ec2api import clients
 from ec2api.db import api as db_api
 from ec2api import exception
 from ec2api.i18n import _
 
 
-Validator = common.Validator
+HOST_TARGET = 'host'
+VPN_TARGET = 'vpn'
+
+
+"""Route tables related API implementation
+"""
+
+
+class Validator(common.Validator):
+
+    def igw_or_vgw_id(self, id):
+        self.ec2_id(id, ['igw', 'vgw'])
 
 
 def create_route_table(context, vpc_id):
@@ -70,6 +83,12 @@ def delete_route(context, route_table_id, destination_cidr_block):
         raise exception.InvalidRouteNotFound(
             route_table_id=route_table_id,
             destination_cidr_block=destination_cidr_block)
+    update_target = _get_route_target(route)
+    if update_target == VPN_TARGET:
+        vpn_gateway = db_api.get_item_by_id(context, route['gateway_id'])
+        if (not vpn_gateway or
+                vpn_gateway['vpc_id'] != route_table['vpc_id']):
+            update_target = None
     rollback_route_table_state = copy.deepcopy(route_table)
     del route_table['routes'][route_index]
     with common.OnCrashCleaner() as cleaner:
@@ -77,12 +96,52 @@ def delete_route(context, route_table_id, destination_cidr_block):
         cleaner.addCleanup(db_api.update_item, context,
                            rollback_route_table_state)
 
-        _update_routes_in_associated_subnets(context, route_table, cleaner,
-                                             rollback_route_table_state)
+        if update_target:
+            _update_routes_in_associated_subnets(
+                context, cleaner, route_table, update_target=update_target)
 
     return True
 
 
+def enable_vgw_route_propagation(context, route_table_id, gateway_id):
+    route_table = ec2utils.get_db_item(context, route_table_id)
+    # NOTE(ft): AWS returns GatewayNotAttached for all invalid cases of
+    # gateway_id value
+    vpn_gateway = ec2utils.get_db_item(context, gateway_id)
+    if vpn_gateway['vpc_id'] != route_table['vpc_id']:
+        raise exception.GatewayNotAttached(gw_id=vpn_gateway['id'],
+                                           vpc_id=route_table['vpc_id'])
+    if vpn_gateway['id'] in route_table.setdefault('propagating_gateways', []):
+        return True
+    with common.OnCrashCleaner() as cleaner:
+        _append_propagation_to_route_table_item(context, route_table,
+                                                vpn_gateway['id'])
+        cleaner.addCleanup(_remove_propagation_from_route_table_item,
+                           context, route_table, vpn_gateway['id'])
+
+        _update_routes_in_associated_subnets(context, cleaner, route_table,
+                                             update_target=VPN_TARGET)
+    return True
+
+
+def disable_vgw_route_propagation(context, route_table_id, gateway_id):
+    route_table = ec2utils.get_db_item(context, route_table_id)
+    if gateway_id not in route_table.get('propagating_gateways', []):
+        return True
+    vpn_gateway = db_api.get_item_by_id(context, gateway_id)
+
+    with common.OnCrashCleaner() as cleaner:
+        _remove_propagation_from_route_table_item(context, route_table,
+                                                  gateway_id)
+        cleaner.addCleanup(_append_propagation_to_route_table_item,
+                           context, route_table, gateway_id)
+
+        if vpn_gateway and vpn_gateway['vpc_id'] == route_table['vpc_id']:
+            _update_routes_in_associated_subnets(context, cleaner, route_table,
+                                                 update_target=VPN_TARGET)
+    return True
+
+
 def associate_route_table(context, route_table_id, subnet_id):
     route_table = ec2utils.get_db_item(context, route_table_id)
     subnet = ec2utils.get_db_item(context, subnet_id)
@@ -98,15 +157,11 @@ def associate_route_table(context, route_table_id, subnet_id):
         msg = msg % {'rtb_id': route_table_id}
         raise exception.ResourceAlreadyAssociated(msg)
 
-    vpc = db_api.get_item_by_id(context, subnet['vpc_id'])
-    main_route_table = db_api.get_item_by_id(context, vpc['route_table_id'])
     with common.OnCrashCleaner() as cleaner:
         _associate_subnet_item(context, subnet, route_table['id'])
         cleaner.addCleanup(_disassociate_subnet_item, context, subnet)
 
-        _update_subnet_host_routes(
-            context, subnet, route_table,
-            cleaner=cleaner, rollback_route_table_object=main_route_table)
+        _update_subnet_routes(context, cleaner, subnet, route_table)
 
     return {'associationId': ec2utils.change_ec2_id_kind(subnet['id'],
                                                          'rtbassoc')}
@@ -119,27 +174,21 @@ def replace_route_table_association(context, association_id, route_table_id):
         vpc = db_api.get_item_by_id(
             context, ec2utils.change_ec2_id_kind(association_id, 'vpc'))
         if vpc is None:
-            raise exception.InvalidAssociationIDNotFound(
-                id=association_id)
+            raise exception.InvalidAssociationIDNotFound(id=association_id)
 
-        rollabck_route_table_object = db_api.get_item_by_id(
-            context, vpc['route_table_id'])
+        rollback_route_table_id = vpc['route_table_id']
         with common.OnCrashCleaner() as cleaner:
             _associate_vpc_item(context, vpc, route_table['id'])
             cleaner.addCleanup(_associate_vpc_item, context, vpc,
-                               rollabck_route_table_object['id'])
+                               rollback_route_table_id)
 
-            # NOTE(ft): this can cause unnecessary update of subnets, which are
-            # associated with the route table
             _update_routes_in_associated_subnets(
-                context, route_table, cleaner,
-                rollabck_route_table_object, is_main=True)
+                context, cleaner, route_table, default_associations_only=True)
     else:
         subnet = db_api.get_item_by_id(
             context, ec2utils.change_ec2_id_kind(association_id, 'subnet'))
         if subnet is None or 'route_table_id' not in subnet:
-            raise exception.InvalidAssociationIDNotFound(
-                id=association_id)
+            raise exception.InvalidAssociationIDNotFound(id=association_id)
         if subnet['vpc_id'] != route_table['vpc_id']:
             msg = _('Route table association %(rtbassoc_id)s and route table '
                     '%(rtb_id)s belong to different networks')
@@ -147,16 +196,13 @@ def replace_route_table_association(context, association_id, route_table_id):
                          'rtb_id': route_table_id}
             raise exception.InvalidParameterValue(msg)
 
-        rollabck_route_table_object = db_api.get_item_by_id(
-            context, subnet['route_table_id'])
+        rollback_route_table_id = subnet['route_table_id']
         with common.OnCrashCleaner() as cleaner:
             _associate_subnet_item(context, subnet, route_table['id'])
             cleaner.addCleanup(_associate_subnet_item, context, subnet,
-                               rollabck_route_table_object['id'])
+                               rollback_route_table_id)
 
-            _update_subnet_host_routes(
-                context, subnet, route_table, cleaner=cleaner,
-                rollback_route_table_object=rollabck_route_table_object)
+            _update_subnet_routes(context, cleaner, subnet, route_table)
 
     return {'newAssociationId': association_id}
 
@@ -168,27 +214,22 @@ def disassociate_route_table(context, association_id):
         vpc = db_api.get_item_by_id(
             context, ec2utils.change_ec2_id_kind(association_id, 'vpc'))
         if vpc is None:
-            raise exception.InvalidAssociationIDNotFound(
-                id=association_id)
+            raise exception.InvalidAssociationIDNotFound(id=association_id)
         msg = _('Cannot disassociate the main route table association '
                 '%(rtbassoc_id)s') % {'rtbassoc_id': association_id}
         raise exception.InvalidParameterValue(msg)
     if 'route_table_id' not in subnet:
-        raise exception.InvalidAssociationIDNotFound(
-            id=association_id)
+        raise exception.InvalidAssociationIDNotFound(id=association_id)
 
-    rollback_route_table_object = db_api.get_item_by_id(
-        context, subnet['route_table_id'])
+    rollback_route_table_id = subnet['route_table_id']
     vpc = db_api.get_item_by_id(context, subnet['vpc_id'])
     main_route_table = db_api.get_item_by_id(context, vpc['route_table_id'])
     with common.OnCrashCleaner() as cleaner:
         _disassociate_subnet_item(context, subnet)
         cleaner.addCleanup(_associate_subnet_item, context, subnet,
-                           rollback_route_table_object['id'])
+                           rollback_route_table_id)
 
-        _update_subnet_host_routes(
-            context, subnet, main_route_table, cleaner=cleaner,
-            rollback_route_table_object=rollback_route_table_object)
+        _update_subnet_routes(context, cleaner, subnet, main_route_table)
 
     return True
 
@@ -226,7 +267,8 @@ class RouteTableDescriber(common.TaggableItemsDescriber,
             is_main=(self.vpcs[route_table['vpc_id']]['route_table_id'] ==
                      route_table['id']),
             gateways=self.gateways,
-            network_interfaces=self.network_interfaces)
+            network_interfaces=self.network_interfaces,
+            vpn_connections_by_gateway_id=self.vpn_connections_by_gateway_id)
 
     def get_db_items(self):
         associations = collections.defaultdict(list)
@@ -236,17 +278,25 @@ class RouteTableDescriber(common.TaggableItemsDescriber,
         self.associations = associations
         vpcs = db_api.get_items(self.context, 'vpc')
         self.vpcs = {vpc['id']: vpc for vpc in vpcs}
-        gateways = db_api.get_items(self.context, 'igw')
-        self.gateways = {igw['id']: igw for igw in gateways}
+        gateways = (db_api.get_items(self.context, 'igw') +
+                    db_api.get_items(self.context, 'vgw'))
+        self.gateways = {gw['id']: gw for gw in gateways}
         # TODO(ft): scan route tables to get only used instances and
         # network interfaces to reduce DB and Nova throughput
         network_interfaces = db_api.get_items(self.context, 'eni')
         self.network_interfaces = {eni['id']: eni
                                    for eni in network_interfaces}
+        vpn_connections = db_api.get_items(self.context, 'vpn')
+        vpns_by_gateway_id = {}
+        for vpn in vpn_connections:
+            vpns = vpns_by_gateway_id.setdefault(vpn['vpn_gateway_id'], [])
+            vpns.append(vpn)
+        self.vpn_connections_by_gateway_id = vpns_by_gateway_id
         return super(RouteTableDescriber, self).get_db_items()
 
 
 def describe_route_tables(context, route_table_id=None, filter=None):
+    ec2utils.check_and_create_default_vpc(context)
     formatted_route_tables = RouteTableDescriber().describe(
             context, ids=route_table_id, filter=filter)
     return {'routeTableSet': formatted_route_tables}
@@ -306,25 +356,30 @@ def _set_route(context, route_table_id, destination_cidr_block,
 
     rollabck_route_table_state = copy.deepcopy(route_table)
     if do_replace:
-        route_count = len(route_table['routes'])
-        route_table['routes'] = [
-            r for r in route_table['routes']
-            if r['destination_cidr_block'] != destination_cidr_block]
-        if route_count == len(route_table['routes']):
+        route_index, old_route = next(
+            ((i, r) for i, r in enumerate(route_table['routes'])
+             if r['destination_cidr_block'] == destination_cidr_block),
+            (None, None))
+        if route_index is None:
             msg = _("There is no route defined for "
                     "'%(destination_cidr_block)s' in the route table. "
                     "Use CreateRoute instead.")
             msg = msg % {'destination_cidr_block': destination_cidr_block}
             raise exception.InvalidParameterValue(msg)
+        else:
+            del route_table['routes'][route_index]
 
     if gateway_id:
         gateway = ec2utils.get_db_item(context, gateway_id)
         if gateway.get('vpc_id') != route_table['vpc_id']:
-            msg = _('Route table %(rtb_id)s and network gateway %(igw_id)s '
-                    'belong to different networks')
-            msg = msg % {'rtb_id': route_table_id,
-                         'igw_id': gateway_id}
-            raise exception.InvalidParameterValue(msg)
+            if ec2utils.get_ec2_id_kind(gateway_id) == 'vgw':
+                raise exception.InvalidGatewayIDNotFound(id=gateway['id'])
+            else:  # igw
+                raise exception.InvalidParameterValue(
+                    _('Route table %(rtb_id)s and network gateway %(igw_id)s '
+                      'belong to different networks') %
+                    {'rtb_id': route_table_id,
+                     'igw_id': gateway_id})
         route = {'gateway_id': gateway['id']}
     elif network_interface_id:
         network_interface = ec2utils.get_db_item(context, network_interface_id)
@@ -358,9 +413,13 @@ def _set_route(context, route_table_id, destination_cidr_block,
         raise exception.InvalidRequest('Parameter VpcPeeringConnectionId is '
                                        'not supported by this implementation')
     route['destination_cidr_block'] = destination_cidr_block
+    update_target = _get_route_target(route)
 
     if do_replace:
         idempotent_call = False
+        old_target = _get_route_target(old_route)
+        if old_target != update_target:
+            update_target = None
     else:
         old_route = next((r for r in route_table['routes']
                           if r['destination_cidr_block'] ==
@@ -377,8 +436,8 @@ def _set_route(context, route_table_id, destination_cidr_block,
         db_api.update_item(context, route_table)
         cleaner.addCleanup(db_api.update_item, context,
                            rollabck_route_table_state)
-        _update_routes_in_associated_subnets(context, route_table, cleaner,
-                                             rollabck_route_table_state)
+        _update_routes_in_associated_subnets(context, cleaner, route_table,
+                                             update_target=update_target)
 
     return True
 
@@ -386,14 +445,20 @@ def _set_route(context, route_table_id, destination_cidr_block,
 def _format_route_table(context, route_table, is_main=False,
                         associated_subnet_ids=[],
                         gateways={},
-                        network_interfaces={}):
+                        network_interfaces={},
+                        vpn_connections_by_gateway_id={}):
     vpc_id = route_table['vpc_id']
-    ec2_route_table = {'routeTableId': route_table['id'],
-                       'vpcId': vpc_id,
-                       'routeSet': [],
-                       # NOTE(ft): AWS returns empty tag set for a route table
-                       # if no tag exists
-                       'tagSet': []}
+    ec2_route_table = {
+        'routeTableId': route_table['id'],
+        'vpcId': vpc_id,
+        'routeSet': [],
+        'propagatingVgwSet': [
+            {'gatewayId': vgw_id}
+            for vgw_id in route_table.get('propagating_gateways', [])],
+        # NOTE(ft): AWS returns empty tag set for a route table
+        # if no tag exists
+        'tagSet': [],
+    }
     # TODO(ft): refactor to get Nova instances outside of this function
     nova = clients.nova(context)
     for route in route_table['routes']:
@@ -437,6 +502,21 @@ def _format_route_table(context, route_table, is_main=False,
                               'state': state})
         ec2_route_table['routeSet'].append(ec2_route)
 
+    for vgw_id in route_table.get('propagating_gateways', []):
+        vgw = gateways.get(vgw_id)
+        if vgw and vgw_id in vpn_connections_by_gateway_id:
+            cidrs = set()
+            vpn_connections = vpn_connections_by_gateway_id[vgw_id]
+            for vpn_connection in vpn_connections:
+                cidrs.update(vpn_connection['cidrs'])
+            state = 'active' if vgw['vpc_id'] == vpc_id else 'blackhole'
+            for cidr in cidrs:
+                ec2_route = {'gatewayId': vgw_id,
+                             'destinationCidrBlock': cidr,
+                             'state': state,
+                             'origin': 'EnableVgwRoutePropagation'}
+                ec2_route_table['routeSet'].append(ec2_route)
+
     associations = []
     if is_main:
         associations.append({
@@ -457,82 +537,124 @@ def _format_route_table(context, route_table, is_main=False,
     return ec2_route_table
 
 
-def _update_routes_in_associated_subnets(context, route_table, cleaner,
-                                         rollabck_route_table_object,
-                                         is_main=None):
-    if is_main is None:
-        vpc = db_api.get_item_by_id(context, route_table['vpc_id'])
-        is_main = vpc['route_table_id'] == route_table['id']
-    if is_main:
-        appropriate_rtb_ids = (route_table['id'], None)
+def _update_routes_in_associated_subnets(context, cleaner, route_table,
+                                         default_associations_only=None,
+                                         update_target=None):
+    if default_associations_only:
+        appropriate_rtb_ids = (None,)
     else:
-        appropriate_rtb_ids = (route_table['id'],)
-    router_objects = _get_router_objects(context, route_table)
+        vpc = db_api.get_item_by_id(context, route_table['vpc_id'])
+        if vpc['route_table_id'] == route_table['id']:
+            appropriate_rtb_ids = (route_table['id'], None)
+        else:
+            appropriate_rtb_ids = (route_table['id'],)
     neutron = clients.neutron(context)
-    for subnet in db_api.get_items(context, 'subnet'):
-        if (subnet['vpc_id'] == route_table['vpc_id'] and
-                subnet.get('route_table_id') in appropriate_rtb_ids):
-            _update_subnet_host_routes(
-                context, subnet, route_table, cleaner=cleaner,
-                rollback_route_table_object=rollabck_route_table_object,
-                router_objects=router_objects, neutron=neutron)
+    subnets = [subnet for subnet in db_api.get_items(context, 'subnet')
+               if (subnet['vpc_id'] == route_table['vpc_id'] and
+                   subnet.get('route_table_id') in appropriate_rtb_ids)]
+    # NOTE(ft): we need to update host routes for both host and vpn target
+    # because vpn-related routes are present in host routes as well
+    _update_host_routes(context, neutron, cleaner, route_table, subnets)
+    if not update_target or update_target == VPN_TARGET:
+        vpn_connection_api._update_vpn_routes(context, neutron, cleaner,
+                                              route_table, subnets)
 
 
-def _update_subnet_host_routes(context, subnet, route_table, cleaner=None,
-                               rollback_route_table_object=None,
-                               router_objects=None, neutron=None):
-    neutron = neutron or clients.neutron(context)
-    os_subnet = neutron.show_subnet(subnet['os_id'])['subnet']
+def _update_subnet_routes(context, cleaner, subnet, route_table):
+    neutron = clients.neutron(context)
+    _update_host_routes(context, neutron, cleaner, route_table, [subnet])
+    vpn_connection_api._update_vpn_routes(context, neutron, cleaner,
+                                          route_table, [subnet])
+
+
+def _update_host_routes(context, neutron, cleaner, route_table, subnets):
+    destinations = _get_active_route_destinations(context, route_table)
+    for subnet in subnets:
+        # TODO(ft): do list subnet w/ filters instead of show one by one
+        os_subnet = neutron.show_subnet(subnet['os_id'])['subnet']
+        host_routes, gateway_ip = _get_subnet_host_routes_and_gateway_ip(
+            context, route_table, os_subnet['cidr'], destinations)
+        neutron.update_subnet(subnet['os_id'],
+                              {'subnet': {'host_routes': host_routes,
+                                          'gateway_ip': gateway_ip}})
+        cleaner.addCleanup(
+            neutron.update_subnet, subnet['os_id'],
+            {'subnet': {'host_routes': os_subnet['host_routes'],
+                        'gateway_ip': os_subnet['gateway_ip']}})
+
+
+def _get_active_route_destinations(context, route_table):
+    vpn_connections = {vpn['vpn_gateway_id']: vpn
+                       for vpn in db_api.get_items(context, 'vpn')}
+    dst_ids = [route[id_key]
+               for route in route_table['routes']
+               for id_key in ('gateway_id', 'network_interface_id')
+               if route.get(id_key) is not None]
+    dst_ids.extend(route_table.get('propagating_gateways', []))
+    destinations = {item['id']: item
+                    for item in db_api.get_items_by_ids(context, dst_ids)
+                    if (item['vpc_id'] == route_table['vpc_id'] and
+                        (ec2utils.get_ec2_id_kind(item['id']) != 'vgw' or
+                         item['id'] in vpn_connections))}
+    for vpn in six.itervalues(vpn_connections):
+        if vpn['vpn_gateway_id'] in destinations:
+            destinations[vpn['vpn_gateway_id']]['vpn_connection'] = vpn
+    return destinations
+
+
+def _get_subnet_host_routes_and_gateway_ip(context, route_table, cidr_block,
+                                           destinations=None):
+    if not destinations:
+        destinations = _get_active_route_destinations(context, route_table)
     gateway_ip = str(netaddr.IPAddress(
-        netaddr.IPNetwork(os_subnet['cidr']).first + 1))
-    host_routes = _get_subnet_host_routes(context, route_table, gateway_ip,
-                                          router_objects)
-    neutron.update_subnet(subnet['os_id'],
-                          {'subnet': {'host_routes': host_routes}})
-    if cleaner and rollback_route_table_object:
-        cleaner.addCleanup(_update_subnet_host_routes, context, subnet,
-                           rollback_route_table_object)
+        netaddr.IPNetwork(cidr_block).first + 1))
 
-
-def _get_router_objects(context, route_table):
-    return dict((route['gateway_id'],
-                 db_api.get_item_by_id(context, route['gateway_id']))
-                if route.get('gateway_id') else
-                (route['network_interface_id'],
-                 db_api.get_item_by_id(context, route['network_interface_id']))
-                for route in route_table['routes']
-                if route.get('gateway_id') or 'network_interface_id' in route)
-
-
-def _get_subnet_host_routes(context, route_table, gateway_ip,
-                            router_objects=None):
     def get_nexthop(route):
         if 'gateway_id' in route:
             gateway_id = route['gateway_id']
-            if gateway_id:
-                gateway = (router_objects[route['gateway_id']]
-                           if router_objects else
-                           db_api.get_item_by_id(context, gateway_id))
-                if (not gateway or
-                        gateway.get('vpc_id') != route_table['vpc_id']):
+            if gateway_id and gateway_id not in destinations:
                     return '127.0.0.1'
             return gateway_ip
-        network_interface = (
-            router_objects[route['network_interface_id']]
-            if router_objects else
-            db_api.get_item_by_id(context, route['network_interface_id']))
+        network_interface = destinations.get(route['network_interface_id'])
         if not network_interface:
             return '127.0.0.1'
         return network_interface['private_ip_address']
 
-    host_routes = [{'destination': route['destination_cidr_block'],
-                    'nexthop': get_nexthop(route)}
-                   for route in route_table['routes']]
-    if not any(r['destination'] == '0.0.0.0/0' for r in host_routes):
-        host_routes.append({'destination': '0.0.0.0/0',
-                            'nexthop': '127.0.0.1'})
+    host_routes = []
+    subnet_gateway_is_used = False
+    for route in route_table['routes']:
+        nexthop = get_nexthop(route)
+        cidr = route['destination_cidr_block']
+        if cidr == '0.0.0.0/0':
+            if nexthop == '127.0.0.1':
+                continue
+            elif nexthop == gateway_ip:
+                subnet_gateway_is_used = True
+        host_routes.append({'destination': cidr,
+                            'nexthop': nexthop})
+    host_routes.extend(
+        {'destination': cidr,
+         'nexthop': gateway_ip}
+        for vgw_id in route_table.get('propagating_gateways', [])
+        for cidr in (destinations.get(vgw_id, {}).get('vpn_connection', {}).
+                     get('cidrs', [])))
 
-    return host_routes
+    if not subnet_gateway_is_used:
+        # NOTE(andrey-mp): add route to metadata server
+        host_routes.append(
+            {'destination': '169.254.169.254/32',
+             'nexthop': gateway_ip})
+        # NOTE(ft): gateway_ip is set to None to allow correct handling
+        # of 0.0.0.0/0 route by Neutron.
+        gateway_ip = None
+    return host_routes, gateway_ip
+
+
+def _get_route_target(route):
+    if ec2utils.get_ec2_id_kind(route.get('gateway_id') or '') == 'vgw':
+        return VPN_TARGET
+    else:
+        return HOST_TARGET
 
 
 def _associate_subnet_item(context, subnet, route_table_id):
@@ -548,3 +670,18 @@ def _disassociate_subnet_item(context, subnet):
 def _associate_vpc_item(context, vpc, route_table_id):
     vpc['route_table_id'] = route_table_id
     db_api.update_item(context, vpc)
+
+
+def _append_propagation_to_route_table_item(context, route_table, gateway_id):
+    vgws = route_table.setdefault('propagating_gateways', [])
+    vgws.append(gateway_id)
+    db_api.update_item(context, route_table)
+
+
+def _remove_propagation_from_route_table_item(context, route_table,
+                                              gateway_id):
+    vgws = route_table['propagating_gateways']
+    vgws.remove(gateway_id)
+    if not vgws:
+        del route_table['propagating_gateways']
+    db_api.update_item(context, route_table)
